use super::{
    safecommandbuffer::{ComputeShaderUnit, SafeCommandBuffer},
    shadercompiler::{Shaders, ShadersType},
};

use anyhow::Result;
use asset::*;
use paste;

use crate::prelude::*;

use std::mem::drop;
use std::sync::{Arc, Mutex};

macro_rules! buffer_handler {
    ( $(($name: ident, $primitive_type: ty),)+ ) => {
        paste::item!{
            pub struct BufferHandler {
                valid: Vec<bool>,

                $(
                    [<$name _renderer>]: SafeCommandBuffer<$primitive_type>,
                    [<$name _iter>]: Option<Box<dyn Fn(
                        &SafeCommandBuffer<$primitive_type>,
                        &mut CommandBufferRecorder<'_>,
                        &PrimitiveData<$primitive_type>,
                        &[&Arc<DescriptorSet>],
                        &View,
                    ) -> Result<()> + Send + Sync>>,
                )+

                transparency_buffer: Option<(Arc<ValidCommandBuffers>, Arc<RenderPass>)>,
            }
        }

        paste::item!{
            impl BufferHandler {
                $(
                    pub fn [<set_ $name _function>](
                        &mut self,
                        f: Box<dyn Fn(
                            &SafeCommandBuffer<$primitive_type>,
                            &mut CommandBufferRecorder<'_>,
                            &PrimitiveData<$primitive_type>,
                            &[&Arc<DescriptorSet>],
                            &View,
                        ) -> Result<()> + Send + Sync>
                    )
                    {
                        self.[<$name _iter>] = Some(f);
                    }
                )+

                pub fn new(
                    create_info: &BufferHandlerCreateInfo<'_>,
                    separate_by_opacity: bool,
                ) -> Result<BufferHandler> {
                    let valid = vec![false; create_info.image_count];

                    Ok(BufferHandler {
                        valid,

                        $(
                            [<$name _renderer>]: match &create_info.safe_command_buffer_type.[<$name _handler_type>] {
                                SafeCommandBufferType::Graphics(func) => {
                                    let (vertex_shader, fragment_shader) = create_info.$name.shaders.graphics();

                                    let mut pipeline_layout_builder = PipelineLayout::builder();

                                    for descriptor_set_layout in create_info.$name.descriptor_layouts.iter() {
                                        pipeline_layout_builder =
                                            pipeline_layout_builder.add_descriptor_set_layout(*descriptor_set_layout);
                                    }

                                    for push_constant in create_info.$name.push_constants.iter() {
                                        pipeline_layout_builder =
                                            pipeline_layout_builder.add_push_constant(push_constant.clone());
                                    }

                                    let pipeline_layout = pipeline_layout_builder.build(create_info.device.clone())?;

                                    let pipeline = func(
                                        create_info.device.clone(),
                                        &pipeline_layout,
                                        vertex_shader.clone(),
                                        fragment_shader.clone(),
                                        create_info.safe_command_buffer_type.render_pass.as_ref().unwrap(),
                                        create_info.safe_command_buffer_type.attachment_info.unwrap(),
                                        create_info.safe_command_buffer_type.sub_pass_index.unwrap()
                                    )?;

                                    SafeCommandBuffer::<$primitive_type>::new_rasterize(
                                        create_info.device,
                                        create_info.queue,
                                        create_info.image_count,
                                        &create_info.safe_command_buffer_type.render_pass.as_ref().unwrap(),
                                        create_info.safe_command_buffer_type.sub_pass_index.unwrap(),
                                        pipeline
                                    )?
                                },
                                SafeCommandBufferType::Compute => SafeCommandBuffer::new_compute(
                                    create_info.device,
                                    create_info.queue,
                                    create_info.image_count,
                                    &[ComputeShaderUnit::new(
                                        &create_info.$name.descriptor_layouts,
                                        create_info.$name.push_constants,
                                        create_info.$name.shaders.compute(),
                                    )],
                                )?,
                            },
                        )+

                        $(
                            [<$name _iter>]: None,
                        )+

                        transparency_buffer: if separate_by_opacity {
                            Some({
                                let command_buffer = ValidCommandBuffers::new(
                                    create_info.device,
                                    create_info.queue,
                                    create_info.image_count,
                                )?;

                                (command_buffer, create_info.safe_command_buffer_type.render_pass.as_ref().unwrap().clone())
                            })
                        } else {
                            None
                        },
                    })

                }

                pub fn invalidate(&mut self) {
                    for valid in &mut self.valid {
                        *valid = false;
                    }

                    $(
                        self.[<$name _renderer>].invalidate();
                    )+
                }

                fn begin(&self, framebuffer: &Arc<Framebuffer>, index: usize) -> Result<BufferRecorder<'_>> {
                    let render_area = VkExtent2D {
                        width: framebuffer.width(),
                        height: framebuffer.height(),
                    };

                    let viewport = [VkViewport {
                        x: 0.0,
                        y: 0.0,
                        width: render_area.width as f32,
                        height: render_area.height as f32,
                        minDepth: 0.0,
                        maxDepth: 1.0,
                    }];

                    let scissor = [VkRect2D {
                        extent: render_area,
                        offset: VkOffset2D { x: 0, y: 0 },
                    }];

                    Ok(BufferRecorder {
                        $(
                            [<$name _recorder>]: {
                                let recorder = self.[<$name _renderer>].begin(Some(framebuffer), index)?;

                                recorder.set_viewport(&viewport);
                                recorder.set_scissor(&scissor);

                                recorder
                            },
                        )+

                        transparency_recorder: {
                            let tmp: Option<Result<CommandBufferRecorder<'_>>> =
                                self.transparency_buffer.as_ref().map(|(buffer, render_pass)| {
                                let (command_buffer, _) = buffer.current(index);

                                let inheritance_info = CommandBuffer::inheritance_info(
                                    Some(&render_pass),
                                    Some(0),
                                    Some(framebuffer),
                                    None,
                                );

                                let mut begin_info = VkCommandBufferBeginInfo::new(
                                    VK_COMMAND_BUFFER_USAGE_RENDER_PASS_CONTINUE_BIT
                                        | VK_COMMAND_BUFFER_USAGE_SIMULTANEOUS_USE_BIT,
                                );

                                begin_info.set_inheritance_info(&inheritance_info);

                                let recorder = command_buffer.begin(begin_info)?;

                                recorder.set_viewport(&viewport);
                                recorder.set_scissor(&scissor);

                                Ok(recorder)
                            });


                            tmp.transpose()?
                        },
                    })
                }

                fn begin_headless(&self, index: usize) -> Result<BufferRecorder<'_>> {
                    Ok(BufferRecorder {
                        $(
                            [<$name _recorder>]: self.[<$name _renderer>].begin_headless(index)?,
                        )+

                        transparency_recorder: self.transparency_buffer.as_ref().map(|(buffer, _render_pass)| {
                            let (command_buffer, _) = buffer.current(index);

                            let inheritance_info = CommandBuffer::inheritance_info(None, None, None, None);

                            let mut begin_info =
                                VkCommandBufferBeginInfo::new(VK_COMMAND_BUFFER_USAGE_SIMULTANEOUS_USE_BIT);

                            begin_info.set_inheritance_info(&inheritance_info);

                            command_buffer.begin(begin_info)
                        }).transpose()?,
                    })
                }

                fn end<'a>(&'a self, buffers: &mut Vec<&'a CommandBuffer>, index: usize, recorders: BufferRecorder<'_>) -> Result<()> {
                    $(
                        {
                            let cmd_buffer = self.[<$name _renderer>].end(recorders.[<$name _recorder>], index)?;

                            if cmd_buffer.calls() != 0 {
                                buffers.push(cmd_buffer);
                            }
                        }
                    )+

                    if let Some(transparency_recorder) = recorders.transparency_recorder {
                        // force drop, ends buffer recording
                        drop(transparency_recorder);

                        let cmd_buffer = self.transparency_buffer.as_ref().unwrap().0.current(index).0;

                        if cmd_buffer.calls() != 0 {
                            buffers.push(cmd_buffer);
                        }
                    }

                    Ok(())
                }

                fn reuse<'a>(&'a self, buffers: &mut Vec<&'a CommandBuffer>, index: usize) -> Result<()> {
                    $(
                        if let Some(command_buffer) = self.[<$name _renderer>].current(index) {
                            if command_buffer.calls() != 0 {
                                buffers.push(command_buffer);
                            }
                        }
                    )+

                    if let Some((buffer, _)) = &self.transparency_buffer {
                        let command_buffer = buffer.current(index).0;

                        if command_buffer.calls() != 0 {
                            buffers.push(command_buffer);
                        }
                    }

                    Ok(())
                }

                fn iterate(
                    &self,
                    draw: &Draw,
                    additional_sets: &[&Arc<DescriptorSet>],
                    recorders: &mut BufferRecorder<'_>,
                    view: &View,
                ) -> Result<()> {
                    for mesh in draw.iter() {
                        for primitive_type in &mesh.primitive_types {
                            match &primitive_type {
                                $(
                                    PrimitiveDataTypes::$primitive_type(primitive) => {
                                        match self.[<$name _iter>] {
                                            Some(ref f) => {
                                                match &mut recorders.transparency_recorder {
                                                    Some(transparency_recorder) => {
                                                        if primitive.opaque {
                                                            f(&self.[<$name _renderer>], &mut recorders.[<$name _recorder>], primitive, additional_sets, view)?;
                                                        } else {
                                                            transparency_recorder.bind_pipeline(self.[<$name _renderer>].current_pipeline())?;

                                                            f(&self.[<$name _renderer>], transparency_recorder, primitive, additional_sets, view)?;
                                                        }
                                                    }
                                                    None => {
                                                        f(&self.[<$name _renderer>], &mut recorders.[<$name _recorder>], primitive, additional_sets, view)?;
                                                    }
                                                }
                                            }
                                            None => {
                                                panic!("BufferHandler: draw function not set for {}", stringify!($primitive_type));
                                            }
                                        }
                                    }
                                )+
                            }
                        }
                    }

                    Ok(())
                }
            }
        }

        paste::item!{
            struct BufferRecorder<'a> {
                $(
                    [<$name _recorder>]: CommandBufferRecorder<'a>,
                )+

                transparency_recorder: Option<CommandBufferRecorder<'a>>,
            }


            pub struct BufferHandlerType<'a> {
                render_pass: Option<Arc<RenderPass>>,
                attachment_info: Option<&'a [bool]>,
                sub_pass_index: Option<u32>,

                $(
                    [<$name _handler_type>]: SafeCommandBufferType,
                    [<$name _pipeline_layout>]: PipelineLayoutInfo<'a>,
                )+
            }

            impl<'a> BufferHandlerType<'a> {
                pub fn graphics<$([<F $name:camel>],)+>(
                    render_pass: &Arc<RenderPass>,
                    attachment_info: &'a [bool],
                    sub_pass_index: u32,
                    $(
                        [<$name _handler_type>]: [<F $name:camel>],
                        [<$name _pipeline_layout>]: PipelineLayoutInfo<'a>,
                    )+
                ) -> Self
                where
                    $(
                        [<F $name:camel>]: Fn(
                            Arc<Device>,
                            &Arc<PipelineLayout>,
                            Arc<ShaderModule<shader_type::Vertex>>,
                            Arc<ShaderModule<shader_type::Fragment>>,
                            &Arc<RenderPass>,
                            &[bool],
                            u32,
                        ) -> Result<Arc<Pipeline>> + Send + Sync + 'static,
                    )+
                {
                    Self {
                        render_pass: Some(render_pass.clone()),
                        attachment_info: Some(attachment_info),
                        sub_pass_index: Some(sub_pass_index),
                        $(
                            [<$name _handler_type>]: SafeCommandBufferType::Graphics(Box::new([<$name _handler_type>])),
                            [<$name _pipeline_layout>],
                        )+
                    }
                }

                pub fn compute($([<$name _pipeline_layout>]: PipelineLayoutInfo<'a>,)+) -> Self {
                    Self {
                        render_pass: None,
                        attachment_info: None,
                        sub_pass_index: None,
                        $(
                            [<$name _handler_type>]: SafeCommandBufferType::Compute,
                            [<$name _pipeline_layout>],
                        )+
                    }
                }
            }


            pub struct BufferHandlerCreateInfo<'a> {
                pub device: &'a Arc<Device>,
                pub queue: &'a Arc<Mutex<Queue>>,
                pub image_count: usize,

                pub safe_command_buffer_type: BufferHandlerType<'a>,

                $(
                    pub $name: SafeCommandBufferCreateInfo<'a>,
                )+
            }

            impl<'a> BufferHandlerCreateInfo<'a> {
                pub fn new(
                    shaders: &'a Shaders,
                    device: &'a Arc<Device>,
                    queue: &'a Arc<Mutex<Queue>>,
                    image_count: usize,
                    buffer_handler_types: BufferHandlerType<'a>
                ) -> Self {
                    BufferHandlerCreateInfo {
                        device,
                        queue,
                        image_count,

                        $(
                            $name: SafeCommandBufferCreateInfo {
                                shaders: &shaders.$name,
                                descriptor_layouts: buffer_handler_types.[<$name _pipeline_layout>].descriptor_layouts.clone(),
                                push_constants: buffer_handler_types.[<$name _pipeline_layout>].push_constants.clone(),
                            },
                        )+

                        safe_command_buffer_type: buffer_handler_types,
                    }
                }
            }
        }
    }
}

buffer_handler!(
    (position, PositionOnly),
    (position_coloruv, PositionColorUV),
    (position_normal, PositionNormal),
    (position_normal_bone, PositionNormalBone),
    (position_normal_color, PositionNormalColor),
    (position_normal_coloruv, PositionNormalColorUV),
    (position_normal_coloruv_bone, PositionNormalColorUVBone),
    (
        position_normal_coloruv_normaluv,
        PositionNormalColorUVNormalUV
    ),
    (
        position_normal_coloruv_normaluv_bone,
        PositionNormalColorUVNormalUVBone
    ),
);

pub enum SafeCommandBufferType {
    Graphics(
        #[allow(clippy::type_complexity)]
        Box<
            dyn Fn(
                Arc<Device>,
                &Arc<PipelineLayout>,
                Arc<ShaderModule<shader_type::Vertex>>,
                Arc<ShaderModule<shader_type::Fragment>>,
                &Arc<RenderPass>,
                &[bool],
                u32,
            ) -> Result<Arc<Pipeline>>,
        >,
    ),
    Compute,
}

pub struct PipelineLayoutInfo<'a> {
    pub descriptor_layouts: Vec<&'a dyn VkHandle<VkDescriptorSetLayout>>,
    pub push_constants: &'a [VkPushConstantRange],
}

pub struct SafeCommandBufferCreateInfo<'a> {
    pub shaders: &'a ShadersType,
    pub descriptor_layouts: Vec<&'a dyn VkHandle<VkDescriptorSetLayout>>,
    pub push_constants: &'a [VkPushConstantRange],
}

impl BufferHandler {
    pub fn execute<'a>(
        &'a mut self,
        framebuffer: Option<&Arc<Framebuffer>>,
        content: &[&EntityObject],
        additional_sets: &[&Arc<DescriptorSet>],
        buffers: &mut Vec<&'a CommandBuffer>,
        index: usize,
        view: &View,
    ) -> Result<()> {
        if !self.valid[index] {
            self.valid[index] = true;

            let mut recorders = match framebuffer {
                Some(fb) => self.begin(fb, index)?,
                None => self.begin_headless(index)?,
            };

            self.iterate_content(content, additional_sets, &mut recorders, view)?;

            self.end(buffers, index, recorders)?;
        } else {
            self.reuse(buffers, index)?;
        }

        Ok(())
    }

    fn iterate_content(
        &self,
        content: &[&EntityObject],
        additional_sets: &[&Arc<DescriptorSet>],
        recorders: &mut BufferRecorder<'_>,
        view: &View,
    ) -> Result<()> {
        content.iter().try_for_each(|entity| {
            if let Ok(draw) = entity.get_component::<Draw>() {
                self.iterate(draw, additional_sets, recorders, view)?;
            }

            Ok(())
        })
    }
}
